@@ -33,7 +33,10 @@ from ..linkers import (
     CompCertDynamicLinker,
     C2000Linker,
     C2000DynamicLinker,
+    TILinker,
+    TIDynamicLinker,
     DLinker,
+    NAGDynamicLinker,
     NvidiaHPC_DynamicLinker,
     PGIDynamicLinker,
     PGIStaticLinker,
@@ -53,6 +56,7 @@ from .c import (
     AppleClangCCompiler,
     ArmCCompiler,
     ArmclangCCompiler,
+    ArmLtdClangCCompiler,
     ClangCCompiler,
     ClangClCCompiler,
     GnuCCompiler,
@@ -66,6 +70,7 @@ from .c import (
     Xc16CCompiler,
     CompCertCCompiler,
     C2000CCompiler,
+    TICCompiler,
     VisualStudioCCompiler,
 )
 from .cpp import (
@@ -73,6 +78,7 @@ from .cpp import (
     AppleClangCPPCompiler,
     ArmCPPCompiler,
     ArmclangCPPCompiler,
+    ArmLtdClangCPPCompiler,
     ClangCPPCompiler,
     ClangClCPPCompiler,
     GnuCPPCompiler,
@@ -84,6 +90,7 @@ from .cpp import (
     PGICPPCompiler,
     CcrxCPPCompiler,
     C2000CPPCompiler,
+    TICPPCompiler,
     VisualStudioCPPCompiler,
 )
 from .cs import MonoCompiler, VisualStudioCsCompiler
@@ -96,6 +103,7 @@ from .d import (
 from .cuda import CudaCompiler
 from .fortran import (
     FortranCompiler,
+    ArmLtdFlangFortranCompiler,
     G95FortranCompiler,
     GnuFortranCompiler,
     ElbrusFortranCompiler,
@@ -123,7 +131,7 @@ from .objcpp import (
     GnuObjCPPCompiler,
 )
 from .cython import CythonCompiler
-from .rust import RustCompiler
+from .rust import RustCompiler, ClippyRustCompiler
 from .swift import SwiftCompiler
 from .vala import ValaCompiler
 from .mixins.visualstudio import VisualStudioLikeCompiler
@@ -141,8 +149,6 @@ import typing as T
 if T.TYPE_CHECKING:
     from ..environment import Environment
     from ..programs import ExternalProgram
-    from .compilers import CompilerType
-
 
 
 # Default compilers and linkers
@@ -167,12 +173,10 @@ if is_windows():
     defaults['cs'] = ['csc', 'mcs']
 else:
     if platform.machine().lower() == 'e2k':
-        # There are no objc or objc++ compilers for Elbrus,
-        # and there's no clang which can build binaries for host.
-        defaults['c'] = ['cc', 'gcc', 'lcc']
-        defaults['cpp'] = ['c++', 'g++', 'l++']
-        defaults['objc'] = []
-        defaults['objcpp'] = []
+        defaults['c'] = ['cc', 'gcc', 'lcc', 'clang']
+        defaults['cpp'] = ['c++', 'g++', 'l++', 'clang++']
+        defaults['objc'] = ['clang']
+        defaults['objcpp'] = ['clang++']
     else:
         defaults['c'] = ['cc', 'gcc', 'clang', 'nvc', 'pgcc', 'icc']
         defaults['cpp'] = ['c++', 'g++', 'clang++', 'nvc++', 'pgc++', 'icpc']
@@ -214,7 +218,7 @@ def compiler_from_language(env: 'Environment', lang: str, for_machine: MachineCh
     }
     return lang_map[lang](env, for_machine) if lang in lang_map else None
 
-def detect_compiler_for(env: 'Environment', lang: str, for_machine: MachineChoice)-> T.Optional[Compiler]:
+def detect_compiler_for(env: 'Environment', lang: str, for_machine: MachineChoice) -> T.Optional[Compiler]:
     comp = compiler_from_language(env, lang, for_machine)
     if comp is not None:
         assert comp.for_machine == for_machine
@@ -239,7 +243,7 @@ def _get_compilers(env: 'Environment', lang: str, for_machine: MachineChoice) ->
         if not env.machines.matches_build_machine(for_machine):
             raise EnvironmentException(f'{lang!r} compiler binary not defined in cross or native file')
         compilers = [[x] for x in defaults[lang]]
-        ccache = BinaryTable.detect_ccache()
+        ccache = BinaryTable.detect_compiler_cache()
 
     if env.machines.matches_build_machine(for_machine):
         exe_wrap: T.Optional[ExternalProgram] = None
@@ -251,8 +255,7 @@ def _get_compilers(env: 'Environment', lang: str, for_machine: MachineChoice) ->
 def _handle_exceptions(
         exceptions: T.Mapping[str, T.Union[Exception, str]],
         binaries: T.List[T.List[str]],
-        bintype: str = 'compiler'
-    ) -> T.NoReturn:
+        bintype: str = 'compiler') -> T.NoReturn:
     errmsg = f'Unknown {bintype}(s): {binaries}'
     if exceptions:
         errmsg += '\nThe following exception(s) were encountered:'
@@ -295,9 +298,11 @@ def detect_static_linker(env: 'Environment', compiler: Compiler) -> StaticLinker
             linkers = default_linkers
     popen_exceptions = {}
     for linker in linkers:
-        if not {'lib', 'lib.exe', 'llvm-lib', 'llvm-lib.exe', 'xilib', 'xilib.exe'}.isdisjoint(linker):
+        linker_name = os.path.basename(linker[0])
+
+        if any(os.path.basename(x) in {'lib', 'lib.exe', 'llvm-lib', 'llvm-lib.exe', 'xilib', 'xilib.exe'} for x in linker):
             arg = '/?'
-        elif not {'ar2000', 'ar2000.exe'}.isdisjoint(linker):
+        elif linker_name in {'ar2000', 'ar2000.exe', 'ar430', 'ar430.exe', 'armar', 'armar.exe'}:
             arg = '?'
         else:
             arg = '--version'
@@ -312,7 +317,7 @@ def detect_static_linker(env: 'Environment', compiler: Compiler) -> StaticLinker
             return VisualStudioLinker(linker, getattr(compiler, 'machine', None))
         if 'ar-Error-Unknown switch: --version' in err:
             return PGIStaticLinker(linker)
-        if p.returncode == 0 and ('armar' in linker or 'armar.exe' in linker):
+        if p.returncode == 0 and 'armar' in linker_name:
             return ArmarLinker(linker)
         if 'DMD32 D Compiler' in out or 'DMD64 D Compiler' in out:
             assert isinstance(compiler, DCompiler)
@@ -323,12 +328,15 @@ def detect_static_linker(env: 'Environment', compiler: Compiler) -> StaticLinker
         if 'GDC' in out and ' based on D ' in out:
             assert isinstance(compiler, DCompiler)
             return DLinker(linker, compiler.arch)
-        if err.startswith('Renesas') and ('rlink' in linker or 'rlink.exe' in linker):
+        if err.startswith('Renesas') and 'rlink' in linker_name:
             return CcrxLinker(linker)
-        if out.startswith('GNU ar') and ('xc16-ar' in linker or 'xc16-ar.exe' in linker):
+        if out.startswith('GNU ar') and 'xc16-ar' in linker_name:
             return Xc16Linker(linker)
-        if out.startswith('TMS320C2000') and ('ar2000' in linker or 'ar2000.exe' in linker):
-            return C2000Linker(linker)
+        if 'Texas Instruments Incorporated' in out:
+            if 'ar2000' in linker_name:
+                return C2000Linker(linker)
+            else:
+                return TILinker(linker)
         if out.startswith('The CompCert'):
             return CompCertLinker(linker)
         if p.returncode == 0:
@@ -342,7 +350,6 @@ def detect_static_linker(env: 'Environment', compiler: Compiler) -> StaticLinker
     _handle_exceptions(popen_exceptions, linkers, 'linker')
 
 
-
 # Compilers
 # =========
 
@@ -367,7 +374,7 @@ def _detect_c_or_cpp_compiler(env: 'Environment', lang: str, for_machine: Machin
             compiler = [compiler]
         compiler_name = os.path.basename(compiler[0])
 
-        if not {'cl', 'cl.exe', 'clang-cl', 'clang-cl.exe'}.isdisjoint(compiler):
+        if any(os.path.basename(x) in {'cl', 'cl.exe', 'clang-cl', 'clang-cl.exe'} for x in compiler):
             # Watcom C provides it's own cl.exe clone that mimics an older
             # version of Microsoft's compiler. Since Watcom's cl.exe is
             # just a wrapper, we skip using it if we detect its presence
@@ -382,9 +389,12 @@ def _detect_c_or_cpp_compiler(env: 'Environment', lang: str, for_machine: Machin
                     return os.path.normcase(os.path.abspath(p))
 
                 watcom_cls = [sanitize(os.path.join(os.environ['WATCOM'], 'BINNT', 'cl')),
-                                sanitize(os.path.join(os.environ['WATCOM'], 'BINNT', 'cl.exe'))]
+                              sanitize(os.path.join(os.environ['WATCOM'], 'BINNT', 'cl.exe')),
+                              sanitize(os.path.join(os.environ['WATCOM'], 'BINNT64', 'cl')),
+                              sanitize(os.path.join(os.environ['WATCOM'], 'BINNT64', 'cl.exe'))]
                 found_cl = sanitize(shutil.which('cl'))
                 if found_cl in watcom_cls:
+                    mlog.debug('Skipping unsupported cl.exe clone at:', found_cl)
                     continue
             arg = '/?'
         elif 'armcc' in compiler_name:
@@ -395,7 +405,8 @@ def _detect_c_or_cpp_compiler(env: 'Environment', lang: str, for_machine: Machin
             arg = '--version'
         elif 'ccomp' in compiler_name:
             arg = '-version'
-        elif 'cl2000' in compiler_name:
+        elif compiler_name in {'cl2000', 'cl2000.exe', 'cl430', 'cl430.exe', 'armcl', 'armcl.exe'}:
+            # TI compiler
             arg = '-version'
         elif compiler_name in {'icl', 'icl.exe'}:
             # if you pass anything to icl you get stuck in a pager
@@ -463,6 +474,20 @@ def _detect_c_or_cpp_compiler(env: 'Environment', lang: str, for_machine: Machin
                 ccache + compiler, version, for_machine, is_cross, info,
                 exe_wrap, linker=linker, full_version=full_version)
 
+        if 'Arm C/C++/Fortran Compiler' in out:
+            arm_ver_match = re.search(r'version (\d+)\.(\d+) \(build number (\d+)\)', out)
+            arm_ver_major = arm_ver_match.group(1)
+            arm_ver_minor = arm_ver_match.group(2)
+            arm_ver_build = arm_ver_match.group(3)
+            version = '.'.join([arm_ver_major, arm_ver_minor, arm_ver_build])
+            if lang == 'c':
+                cls = ArmLtdClangCCompiler
+            elif lang == 'cpp':
+                cls = ArmLtdClangCPPCompiler
+            linker = guess_nix_linker(env, compiler, cls, for_machine)
+            return cls(
+                ccache + compiler, version, for_machine, is_cross, info,
+                exe_wrap, linker=linker)
         if 'armclang' in out:
             # The compiler version is not present in the first line of output,
             # instead it is present in second line, startswith 'Component:'.
@@ -558,8 +583,13 @@ def _detect_c_or_cpp_compiler(env: 'Environment', lang: str, for_machine: Machin
                 raise EnvironmentException(m)
             cls = VisualStudioCCompiler if lang == 'c' else VisualStudioCPPCompiler
             linker = guess_win_linker(env, ['link'], cls, for_machine)
+            # As of this writing, CCache does not support MSVC but sccache does.
+            if 'sccache' in ccache:
+                final_compiler = ccache + compiler
+            else:
+                final_compiler = compiler
             return cls(
-                compiler, version, for_machine, is_cross, info, target,
+                final_compiler, version, for_machine, is_cross, info, target,
                 exe_wrap, full_version=cl_signature, linker=linker)
         if 'PGI Compilers' in out:
             cls = PGICCompiler if lang == 'c' else PGICPPCompiler
@@ -581,6 +611,20 @@ def _detect_c_or_cpp_compiler(env: 'Environment', lang: str, for_machine: Machin
             return cls(
                 ccache + compiler, version, for_machine, is_cross, info,
                 exe_wrap, full_version=full_version, linker=l)
+        if 'TMS320C2000 C/C++' in out or 'MSP430 C/C++' in out or 'TI ARM C/C++ Compiler' in out:
+            lnk: T.Union[T.Type[C2000DynamicLinker], T.Type[TIDynamicLinker]]
+            if 'TMS320C2000 C/C++' in out:
+                cls = C2000CCompiler if lang == 'c' else C2000CPPCompiler
+                lnk = C2000DynamicLinker
+            else:
+                cls = TICCompiler if lang == 'c' else TICPPCompiler
+                lnk = TIDynamicLinker
+
+            env.coredata.add_lang_args(cls.language, cls, for_machine, env)
+            linker = lnk(compiler, for_machine, version=version)
+            return cls(
+                ccache + compiler, version, for_machine, is_cross, info,
+                exe_wrap, full_version=full_version, linker=linker)
         if 'ARM' in out:
             cls = ArmCCompiler if lang == 'c' else ArmCPPCompiler
             env.coredata.add_lang_args(cls.language, cls, for_machine, env)
@@ -612,15 +656,6 @@ def _detect_c_or_cpp_compiler(env: 'Environment', lang: str, for_machine: Machin
                 ccache + compiler, version, for_machine, is_cross, info,
                 exe_wrap, full_version=full_version, linker=linker)
 
-        if 'TMS320C2000 C/C++' in out:
-            cls = C2000CCompiler if lang == 'c' else C2000CPPCompiler
-            env.coredata.add_lang_args(cls.language, cls, for_machine, env)
-            linker = C2000DynamicLinker(compiler, for_machine, version=version)
-            return cls(
-                ccache + compiler, version, for_machine, is_cross, info,
-                exe_wrap, full_version=full_version, linker=linker)
-
-
     _handle_exceptions(popen_exceptions, compilers)
     raise EnvironmentException(f'Unknown compiler {compilers}')
 
@@ -696,15 +731,29 @@ def detect_fortran_compiler(env: 'Environment', for_machine: MachineChoice) -> C
                 if guess_gcc_or_lcc == 'lcc':
                     version = _get_lcc_version_from_defines(defines)
                     cls = ElbrusFortranCompiler
+                    linker = guess_nix_linker(env, compiler, cls, for_machine)
+                    return cls(
+                        compiler, version, for_machine, is_cross, info,
+                        exe_wrap, defines, full_version=full_version, linker=linker)
                 else:
                     version = _get_gnu_version_from_defines(defines)
                     cls = GnuFortranCompiler
+                    linker = guess_nix_linker(env, compiler, cls, for_machine)
+                    return cls(
+                        compiler, version, for_machine, is_cross, info,
+                        exe_wrap, defines, full_version=full_version, linker=linker)
+
+            if 'Arm C/C++/Fortran Compiler' in out:
+                cls = ArmLtdFlangFortranCompiler
+                arm_ver_match = re.search(r'version (\d+)\.(\d+) \(build number (\d+)\)', out)
+                arm_ver_major = arm_ver_match.group(1)
+                arm_ver_minor = arm_ver_match.group(2)
+                arm_ver_build = arm_ver_match.group(3)
+                version = '.'.join([arm_ver_major, arm_ver_minor, arm_ver_build])
                 linker = guess_nix_linker(env, compiler, cls, for_machine)
                 return cls(
-                    compiler, version, for_machine, is_cross, info,
-                    exe_wrap, defines, full_version=full_version,
-                    linker=linker)
-
+                    ccache + compiler, version, for_machine, is_cross, info,
+                    exe_wrap, linker=linker)
             if 'G95' in out:
                 cls = G95FortranCompiler
                 linker = guess_nix_linker(env, compiler, cls, for_machine)
@@ -745,7 +794,7 @@ def detect_fortran_compiler(env: 'Environment', for_machine: MachineChoice) -> C
                 cls = PGIFortranCompiler
                 env.coredata.add_lang_args(cls.language, cls, for_machine, env)
                 linker = PGIDynamicLinker(compiler, for_machine,
-                                            cls.LINKER_PREFIX, [], version=version)
+                                          cls.LINKER_PREFIX, [], version=version)
                 return cls(
                     compiler, version, for_machine, is_cross, info, exe_wrap,
                     full_version=full_version, linker=linker)
@@ -754,29 +803,34 @@ def detect_fortran_compiler(env: 'Environment', for_machine: MachineChoice) -> C
                 cls = NvidiaHPC_FortranCompiler
                 env.coredata.add_lang_args(cls.language, cls, for_machine, env)
                 linker = PGIDynamicLinker(compiler, for_machine,
-                                            cls.LINKER_PREFIX, [], version=version)
+                                          cls.LINKER_PREFIX, [], version=version)
                 return cls(
                     compiler, version, for_machine, is_cross, info, exe_wrap,
                     full_version=full_version, linker=linker)
 
             if 'flang' in out or 'clang' in out:
                 linker = guess_nix_linker(env,
-                    compiler, FlangFortranCompiler, for_machine)
+                                          compiler, FlangFortranCompiler, for_machine)
                 return FlangFortranCompiler(
                     compiler, version, for_machine, is_cross, info,
                     exe_wrap, full_version=full_version, linker=linker)
 
             if 'Open64 Compiler Suite' in err:
                 linker = guess_nix_linker(env,
-                    compiler, Open64FortranCompiler, for_machine)
+                                          compiler, Open64FortranCompiler, for_machine)
                 return Open64FortranCompiler(
                     compiler, version, for_machine, is_cross, info,
                     exe_wrap, full_version=full_version, linker=linker)
 
             if 'NAG Fortran' in err:
-                linker = guess_nix_linker(env,
-                    compiler, NAGFortranCompiler, for_machine)
-                return NAGFortranCompiler(
+                full_version = err.split('\n', 1)[0]
+                version = full_version.split()[-1]
+                cls = NAGFortranCompiler
+                env.coredata.add_lang_args(cls.language, cls, for_machine, env)
+                linker = NAGDynamicLinker(
+                    compiler, for_machine, cls.LINKER_PREFIX, [],
+                    version=version)
+                return cls(
                     compiler, version, for_machine, is_cross, info,
                     exe_wrap, full_version=full_version, linker=linker)
 
@@ -947,6 +1001,13 @@ def detect_rust_compiler(env: 'Environment', for_machine: MachineChoice) -> Rust
             continue
 
         version = search_version(out)
+        cls: T.Type[RustCompiler] = RustCompiler
+
+        # Clippy is a wrapper around rustc, but it doesn't have rustc in it's
+        # output. We can otherwise treat it as rustc.
+        if 'clippy' in out:
+            out = 'rustc'
+            cls = ClippyRustCompiler
 
         if 'rustc' in out:
             # On Linux and mac rustc will invoke gcc (clang for mac
@@ -965,19 +1026,21 @@ def detect_rust_compiler(env: 'Environment', for_machine: MachineChoice) -> Rust
                     'or use the RUST_LD environment variable, otherwise meson '
                     'will override your selection.')
 
+            compiler = compiler.copy()  # avoid mutating the original list
+
             if override is None:
                 extra_args: T.Dict[str, T.Union[str, bool]] = {}
                 always_args: T.List[str] = []
                 if is_link_exe:
-                    compiler.extend(RustCompiler.use_linker_args(cc.linker.exelist[0]))
+                    compiler.extend(cls.use_linker_args(cc.linker.exelist[0]))
                     extra_args['direct'] = True
                     extra_args['machine'] = cc.linker.machine
                 else:
-                    exelist = cc.linker.exelist.copy()
+                    exelist = cc.linker.exelist + cc.linker.get_always_args()
                     if 'ccache' in exelist[0]:
                         del exelist[0]
                     c = exelist.pop(0)
-                    compiler.extend(RustCompiler.use_linker_args(c))
+                    compiler.extend(cls.use_linker_args(c))
 
                     # Also ensure that we pass any extra arguments to the linker
                     for l in exelist:
@@ -988,19 +1051,19 @@ def detect_rust_compiler(env: 'Environment', for_machine: MachineChoice) -> Rust
                 # TODO rewrite this without type: ignore
                 if is_link_exe:
                     linker = type(cc.linker)(for_machine, always_args, exelist=cc.linker.exelist,   # type: ignore
-                                                version=cc.linker.version, **extra_args)            # type: ignore
+                                             version=cc.linker.version, **extra_args)               # type: ignore
                 else:
                     linker = type(cc.linker)(compiler, for_machine, cc.LINKER_PREFIX,
-                                                always_args=always_args, version=cc.linker.version,
-                                                **extra_args) # type: ignore
+                                             always_args=always_args, version=cc.linker.version,
+                                             **extra_args)
             elif 'link' in override[0]:
                 linker = guess_win_linker(env,
-                    override, RustCompiler, for_machine, use_linker_prefix=False)
+                                          override, cls, for_machine, use_linker_prefix=False)
                 # rustc takes linker arguments without a prefix, and
                 # inserts the correct prefix itself.
                 assert isinstance(linker, VisualStudioLikeLinkerMixin)
                 linker.direct = True
-                compiler.extend(RustCompiler.use_linker_args(linker.exelist[0]))
+                compiler.extend(cls.use_linker_args(linker.exelist[0]))
             else:
                 # On linux and macos rust will invoke the c compiler for
                 # linking, on windows it will use lld-link or link.exe.
@@ -1012,10 +1075,10 @@ def detect_rust_compiler(env: 'Environment', for_machine: MachineChoice) -> Rust
                 # Of course, we're not going to use any of that, we just
                 # need it to get the proper arguments to pass to rustc
                 c = linker.exelist[1] if linker.exelist[0].endswith('ccache') else linker.exelist[0]
-                compiler.extend(RustCompiler.use_linker_args(c))
+                compiler.extend(cls.use_linker_args(c))
 
-            env.coredata.add_lang_args(RustCompiler.language, RustCompiler, for_machine, env)
-            return RustCompiler(
+            env.coredata.add_lang_args(cls.language, cls, for_machine, env)
+            return cls(
                 compiler, version, for_machine, is_cross, info, exe_wrap,
                 linker=linker)
 
@@ -1070,17 +1133,17 @@ def detect_d_compiler(env: 'Environment', for_machine: MachineChoice) -> Compile
                 if info.is_windows() or info.is_cygwin():
                     objfile = os.path.basename(f)[:-1] + 'obj'
                     linker = guess_win_linker(env,
-                        exelist,
-                        LLVMDCompiler, for_machine,
-                        use_linker_prefix=True, invoked_directly=False,
-                        extra_args=[f])
+                                              exelist,
+                                              LLVMDCompiler, for_machine,
+                                              use_linker_prefix=True, invoked_directly=False,
+                                              extra_args=[f])
                 else:
                     # LDC writes an object file to the current working directory.
                     # Clean it up.
                     objfile = os.path.basename(f)[:-1] + 'o'
                     linker = guess_nix_linker(env,
-                        exelist, LLVMDCompiler, for_machine,
-                        extra_args=[f])
+                                              exelist, LLVMDCompiler, for_machine,
+                                              extra_args=[f])
             finally:
                 windows_proof_rm(f)
                 windows_proof_rm(objfile)
@@ -1109,13 +1172,13 @@ def detect_d_compiler(env: 'Environment', for_machine: MachineChoice) -> Compile
                 if info.is_windows() or info.is_cygwin():
                     objfile = os.path.basename(f)[:-1] + 'obj'
                     linker = guess_win_linker(env,
-                        exelist, DmdDCompiler, for_machine,
-                        invoked_directly=False, extra_args=[f, arch_arg])
+                                              exelist, DmdDCompiler, for_machine,
+                                              invoked_directly=False, extra_args=[f, arch_arg])
                 else:
                     objfile = os.path.basename(f)[:-1] + 'o'
                     linker = guess_nix_linker(env,
-                        exelist, DmdDCompiler, for_machine,
-                        extra_args=[f, arch_arg])
+                                              exelist, DmdDCompiler, for_machine,
+                                              extra_args=[f, arch_arg])
             finally:
                 windows_proof_rm(f)
                 windows_proof_rm(objfile)
@@ -1145,8 +1208,8 @@ def detect_swift_compiler(env: 'Environment', for_machine: MachineChoice) -> Com
         # As for 5.0.1 swiftc *requires* a file to check the linker:
         with tempfile.NamedTemporaryFile(suffix='.swift') as f:
             linker = guess_nix_linker(env,
-                exelist, SwiftCompiler, for_machine,
-                extra_args=[f.name])
+                                      exelist, SwiftCompiler, for_machine,
+                                      extra_args=[f.name])
         return SwiftCompiler(
             exelist, version, for_machine, is_cross, info, linker=linker)
 
