@@ -22,10 +22,10 @@ from mesonbuild import mlog, build
 from mesonbuild.coredata import MesonException
 from . import ModuleReturnValue
 from . import ExtensionModule
-from . import get_include_args
-from ..dependencies import Dependency, InternalDependency, ExternalProgram
+from ..dependencies import Dependency, InternalDependency
 from ..interpreterbase import FeatureNew, InvalidArguments, noPosargs, noKwargs
 from ..interpreter import CustomTargetHolder
+from ..programs import ExternalProgram
 
 
 def ensure_list(value):
@@ -100,20 +100,19 @@ class HotdocTargetBuilder:
             # When an option expects a single value, the unambiguous way
             # to specify it is with =
             if isinstance(value, str):
-                self.cmd.extend(['%s=%s' % (option, value)])
+                self.cmd.extend([f'{option}={value}'])
             else:
                 self.cmd.extend([option, value])
 
     def check_extra_arg_type(self, arg, value):
-        value = getattr(value, 'held_object', value)
         if isinstance(value, list):
             for v in value:
                 self.check_extra_arg_type(arg, v)
             return
 
-        valid_types = (str, bool, mesonlib.File, build.IncludeDirs, build.CustomTarget, build.BuildTarget)
+        valid_types = (str, bool, mesonlib.File, build.IncludeDirs, build.CustomTarget, build.CustomTargetIndex, build.BuildTarget)
         if not isinstance(value, valid_types):
-            raise InvalidArguments('Argument "%s=%s" should be of type: %s.' % (
+            raise InvalidArguments('Argument "{}={}" should be of type: {}.'.format(
                 arg, value, [t.__name__ for t in valid_types]))
 
     def process_extra_args(self):
@@ -136,12 +135,11 @@ class HotdocTargetBuilder:
                     if force_list and not isinstance(value, list):
                         return [value], uvalue
                     return value, uvalue
-            raise MesonException("%s field value %s is not valid,"
-                                 " valid types are %s" % (argname, value,
-                                                          types))
+            raise MesonException(f"{argname} field value {value} is not valid,"
+                                 f" valid types are {types}")
         except KeyError:
             if mandatory:
-                raise MesonException("%s mandatory field not found" % argname)
+                raise MesonException(f"{argname} mandatory field not found")
 
             if default is not None:
                 return default, default
@@ -188,9 +186,8 @@ class HotdocTargetBuilder:
     def process_dependencies(self, deps):
         cflags = set()
         for dep in mesonlib.listify(ensure_list(deps)):
-            dep = getattr(dep, "held_object", dep)
             if isinstance(dep, InternalDependency):
-                inc_args = get_include_args(dep.include_directories)
+                inc_args = self.state.get_include_args(dep.include_directories)
                 cflags.update([self.replace_dirs_in_string(x)
                                for x in inc_args])
                 cflags.update(self.process_dependencies(dep.libraries))
@@ -212,6 +209,8 @@ class HotdocTargetBuilder:
                 self.add_extension_paths(dep.extra_extension_paths)
             elif isinstance(dep, build.CustomTarget) or isinstance(dep, build.BuildTarget):
                 self._dependencies.append(dep)
+            elif isinstance(dep, build.CustomTargetIndex):
+                self._dependencies.append(dep.target)
 
         return [f.strip('-I') for f in cflags]
 
@@ -232,7 +231,6 @@ class HotdocTargetBuilder:
     def flatten_config_command(self):
         cmd = []
         for arg in mesonlib.listify(self.cmd, flatten=True):
-            arg = getattr(arg, 'held_object', arg)
             if isinstance(arg, mesonlib.File):
                 arg = arg.absolute_path(self.state.environment.get_source_dir(),
                                         self.state.environment.get_build_dir())
@@ -242,9 +240,12 @@ class HotdocTargetBuilder:
                     cmd.append(os.path.join(self.builddir, arg.get_curdir(), inc_dir))
 
                 continue
-            elif isinstance(arg, build.CustomTarget) or isinstance(arg, build.BuildTarget):
+            elif isinstance(arg, (build.BuildTarget, build.CustomTarget)):
                 self._dependencies.append(arg)
                 arg = self.interpreter.backend.get_target_filename_abs(arg)
+            elif isinstance(arg, build.CustomTargetIndex):
+                self._dependencies.append(arg.target)
+                arg = self.interpreter.backend.get_target_filename_abs(arg)
 
             cmd.append(arg)
 
@@ -265,7 +266,7 @@ class HotdocTargetBuilder:
                 res.append(self.ensure_file(val))
             return res
 
-        if not isinstance(value, mesonlib.File):
+        if isinstance(value, str):
             return mesonlib.File.from_source_file(self.sourcedir, self.subdir, value)
 
         return value
@@ -277,21 +278,21 @@ class HotdocTargetBuilder:
             _dir = os.path.join(self.sourcedir, self.subdir, value)
 
         if not os.path.isdir(_dir):
-            raise InvalidArguments('"%s" is not a directory.' % _dir)
+            raise InvalidArguments(f'"{_dir}" is not a directory.')
 
         return os.path.relpath(_dir, os.path.join(self.builddir, self.subdir))
 
-    def check_forbiden_args(self):
+    def check_forbidden_args(self):
         for arg in ['conf_file']:
             if arg in self.kwargs:
-                raise InvalidArguments('Argument "%s" is forbidden.' % arg)
+                raise InvalidArguments(f'Argument "{arg}" is forbidden.')
 
     def add_include_path(self, path):
         self.include_paths[path] = path
 
     def make_targets(self):
-        self.check_forbiden_args()
-        file_types = (str, mesonlib.File)
+        self.check_forbidden_args()
+        file_types = (str, mesonlib.File, build.CustomTarget, build.CustomTargetIndex)
         self.process_known_arg("--index", file_types, mandatory=True, value_processor=self.ensure_file)
         self.process_known_arg("--project-version", str, mandatory=True)
         self.process_known_arg("--sitemap", file_types, mandatory=True, value_processor=self.ensure_file)
@@ -313,7 +314,7 @@ class HotdocTargetBuilder:
         hotdoc_config_name = fullname + '.json'
         hotdoc_config_path = os.path.join(
             self.builddir, self.subdir, hotdoc_config_name)
-        with open(hotdoc_config_path, 'w') as f:
+        with open(hotdoc_config_path, 'w', encoding='utf-8') as f:
             f.write('{}')
 
         self.cmd += ['--conf-file', hotdoc_config_path]
@@ -326,7 +327,7 @@ class HotdocTargetBuilder:
         for path in self.include_paths.keys():
             self.cmd.extend(['--include-path', path])
 
-        if self.state.environment.coredata.get_builtin_option('werror', self.state.subproject):
+        if self.state.environment.coredata.get_option(mesonlib.OptionKey('werror', subproject=self.state.subproject)):
             self.cmd.append('--fatal-warning')
         self.generate_hotdoc_config()
 
@@ -343,20 +344,22 @@ class HotdocTargetBuilder:
                               extra_assets=self._extra_assets,
                               subprojects=self._subprojects,
                               command=target_cmd,
-                              depends=self._dependencies,
-                              output=fullname,
+                              extra_depends=self._dependencies,
+                              outputs=[fullname],
+                              sources=[],
                               depfile=os.path.basename(depfile),
                               build_by_default=self.build_by_default)
 
         install_script = None
         if install is True:
-            install_script = HotdocRunScript(self.build_command, [
+            install_script = self.state.backend.get_executable_serialisation(self.build_command + [
                 "--internal", "hotdoc",
                 "--install", os.path.join(fullname, 'html'),
                 '--name', self.name,
                 '--builddir', os.path.join(self.builddir, self.subdir)] +
                 self.hotdoc.get_command() +
                 ['run', '--conf-file', hotdoc_config_name])
+            install_script.tag = 'doc'
 
         return (target, install_script)
 
@@ -371,13 +374,13 @@ class HotdocTargetHolder(CustomTargetHolder):
     def config_path_method(self, *args, **kwargs):
         conf = self.held_object.hotdoc_conf.absolute_path(self.interpreter.environment.source_dir,
                                                           self.interpreter.environment.build_dir)
-        return self.interpreter.holderify(conf)
+        return conf
 
 
 class HotdocTarget(build.CustomTarget):
     def __init__(self, name, subdir, subproject, hotdoc_conf, extra_extension_paths, extra_assets,
                  subprojects, **kwargs):
-        super().__init__(name, subdir, subproject, kwargs, absolute_paths=True)
+        super().__init__(name, subdir, subproject, **kwargs, absolute_paths=True)
         self.hotdoc_conf = hotdoc_conf
         self.extra_extension_paths = extra_extension_paths
         self.extra_assets = extra_assets
@@ -391,11 +394,6 @@ class HotdocTarget(build.CustomTarget):
         return res
 
 
-class HotdocRunScript(build.RunScript):
-    def __init__(self, script, args):
-        super().__init__(script, args)
-
-
 class HotDocModule(ExtensionModule):
     @FeatureNew('Hotdoc Module', '0.48.0')
     def __init__(self, interpreter):
@@ -408,13 +406,15 @@ class HotDocModule(ExtensionModule):
             from hotdoc.run_hotdoc import run  # noqa: F401
             self.hotdoc.run_hotdoc = run
         except Exception as e:
-            raise MesonException('hotdoc %s required but not found. (%s)' % (
-                MIN_HOTDOC_VERSION, e))
+            raise MesonException(f'hotdoc {MIN_HOTDOC_VERSION} required but not found. ({e})')
+        self.methods.update({
+            'has_extensions': self.has_extensions,
+            'generate_doc': self.generate_doc,
+        })
 
     @noKwargs
     def has_extensions(self, state, args, kwargs):
-        res = self.hotdoc.run_hotdoc(['--has-extension=%s' % extension for extension in args]) == 0
-        return ModuleReturnValue(res, [res])
+        return self.hotdoc.run_hotdoc([f'--has-extension={extension}' for extension in args]) == 0
 
     def generate_doc(self, state, args, kwargs):
         if len(args) != 1:
@@ -424,7 +424,7 @@ class HotDocModule(ExtensionModule):
         project_name = args[0]
         builder = HotdocTargetBuilder(project_name, state, self.hotdoc, self.interpreter, kwargs)
         target, install_script = builder.make_targets()
-        targets = [HotdocTargetHolder(target, self.interpreter)]
+        targets = [target]
         if install_script:
             targets.append(install_script)
 
@@ -432,4 +432,6 @@ class HotDocModule(ExtensionModule):
 
 
 def initialize(interpreter):
-    return HotDocModule(interpreter)
+    mod = HotDocModule(interpreter)
+    mod.interpreter.append_holder_map(HotdocTarget, HotdocTargetHolder)
+    return mod
