@@ -25,8 +25,10 @@ import re
 import subprocess
 import sys
 import textwrap
+import json
 import typing as T
 from pathlib import Path
+from urllib.request import urlopen
 
 PathLike = T.Union[Path,str]
 
@@ -35,10 +37,27 @@ def _get_meson_output(root_dir: Path, args: T.List) -> str:
     env['COLUMNS'] = '80'
     return subprocess.run([str(sys.executable), str(root_dir/'meson.py')] + args, check=True, capture_output=True, text=True, env=env).stdout.strip()
 
+def get_commands(help_output: str) -> T.Set[str]:
+    # Python's argument parser might put the command list to its own line. Or it might not.
+    assert(help_output.startswith('usage: '))
+    lines = help_output.split('\n')
+    line1 = lines[0]
+    line2 = lines[1]
+    if '{' in line1:
+        cmndline = line1
+    else:
+        assert('{' in line2)
+        cmndline = line2
+    cmndstr = cmndline.split('{')[1]
+    assert('}' in cmndstr)
+    help_commands = set(cmndstr.split('}')[0].split(','))
+    assert(len(help_commands) > 0)
+    return {c.strip() for c in help_commands}
+
 def get_commands_data(root_dir: Path) -> T.Dict[str, T.Any]:
     usage_start_pattern = re.compile(r'^usage: ', re.MULTILINE)
     positional_start_pattern = re.compile(r'^positional arguments:[\t ]*[\r\n]+', re.MULTILINE)
-    options_start_pattern = re.compile(r'^optional arguments:[\t ]*[\r\n]+', re.MULTILINE)
+    options_start_pattern = re.compile(r'^(optional arguments|options):[\t ]*[\r\n]+', re.MULTILINE)
     commands_start_pattern = re.compile(r'^[A-Za-z ]*[Cc]ommands:[\t ]*[\r\n]+', re.MULTILINE)
 
     def get_next_start(iterators: T.Sequence[T.Any], end: T.Optional[int]) -> int:
@@ -94,7 +113,7 @@ def get_commands_data(root_dir: Path) -> T.Dict[str, T.Any]:
         return out
 
     output = _get_meson_output(root_dir, ['--help'])
-    commands = set(c.strip() for c in re.findall(r'usage:(?:.+)?{((?:[a-z]+,*)+?)}', output, re.MULTILINE|re.DOTALL)[0].split(','))
+    commands = get_commands(output)
     commands.remove('help')
 
     cmd_data = dict()
@@ -112,9 +131,25 @@ def generate_hotdoc_includes(root_dir: Path, output_dir: Path) -> None:
 
     for cmd, parsed in cmd_data.items():
         for typ in parsed.keys():
-            with open(output_dir / (cmd+'_'+typ+'.inc'), 'w') as f:
+            with open(output_dir / (cmd+'_'+typ+'.inc'), 'w', encoding='utf-8') as f:
                 f.write(parsed[typ])
 
+def generate_wrapdb_table(output_dir: Path) -> None:
+    url = urlopen('https://wrapdb.mesonbuild.com/v2/releases.json')
+    releases = json.loads(url.read().decode())
+    with open(output_dir / 'wrapdb-table.md', 'w', encoding='utf-8') as f:
+        f.write('| Project | Versions | Provided dependencies | Provided programs |\n')
+        f.write('| ------- | -------- | --------------------- | ----------------- |\n')
+        for name, info in releases.items():
+            versions = [f'[{v}](https://wrapdb.mesonbuild.com/v2/{name}_{v}/{name}.wrap)' for v in info['versions']]
+            # Highlight latest version.
+            versions_str = f'<big>**{versions[0]}**</big><br/>' + ', '.join(versions[1:])
+            dependency_names = info.get('dependency_names', [])
+            dependency_names_str = ', '.join(dependency_names)
+            program_names = info.get('program_names', [])
+            program_names_str = ', '.join(program_names)
+            f.write(f'| {name} | {versions_str} | {dependency_names_str} | {program_names_str} |\n')
+
 def regenerate_docs(output_dir: PathLike,
                     dummy_output_file: T.Optional[PathLike]) -> None:
     if not output_dir:
@@ -126,9 +161,10 @@ def regenerate_docs(output_dir: PathLike,
     root_dir = Path(__file__).resolve().parent.parent
 
     generate_hotdoc_includes(root_dir, output_dir)
+    generate_wrapdb_table(output_dir)
 
     if dummy_output_file:
-        with open(output_dir/dummy_output_file, 'w') as f:
+        with open(output_dir/dummy_output_file, 'w', encoding='utf-8') as f:
             f.write('dummy file for custom_target output')
 
 if __name__ == '__main__':
